; NeoGeo CD CDDA track player (2) with visualisation for AS
; http://neogeodev.wikkii.com/
; WTFPL 04/2011 furrtek

    cpu 68000
    supmode on
    
    INCLUDE "regdefs.asm"

NGH          equ   $0000
CDDA         equ   $02
DEBUGDIPS    equ   $10F000

    org RAMSTART
FLAGS:          ds.b 1
VBLC:           ds.w 1          ; frame counter
INDX:           ds.w 1          ; current buffer index
PALOFFS:        ds.w 1          ; palette rotation offset
PALINDX:        ds.w 1          ; current palette index
NEWLEFTLEVELS:  ds.b 128        ; update buffer
OLDLEFTLEVELS:  ds.b 128        ; display buffer

    INCLUDE "cdheader.asm"

; ---- Interrupt handlers ----

VBlank:
    btst    #7,BIOS_SYSSTAT
    bne     .startdone
    jmp     BIOSF_BOOTSCR
.startdone:
    movem.l d0-d7/a0-a6,-(a7)
    move.w  #4,REG_IRQACK
    move.b  d0,REG_DIPSW        ; kick watchdog
    btst.b  #0,FLAGS
    beq     noanim              ; do nothing until main init done
    
    clr.w   INDX                ; start each new frame with buffer index 0

    lea     NEWLEFTLEVELS,a0    ; copy new buffer to display buffer
    lea     OLDLEFTLEVELS,a1
    move.l  #64-1,d0
.copybuffer:
    move.w  (a0)+,(a1)+
    dbra    d0,.copybuffer

    move.w  #FIXMAP+(25+(9*32)),VRAM_ADDR
    lea     VBLC,a0
    jsr     WriteHex            ; display frame counter

    lea     OLDLEFTLEVELS,a0
    move.l  #1,d2               ; first sprite number to use
    move.w  #1,VRAM_MOD
    move.l  #64-1,d7
    clr.l   d4                  ; D4 will be the average of all values

.processlevels:
    clr.l   d1
    move.w  (a0)+,d1
    jsr     convert             ; read and convert level value to binary
    add.l   d1,d4

    move.w  d1,d0

    move.l  d2,d3
    addi.w  #SCB2,d3
    move.w  d3,VRAM_ADDR        ; compute address for current sprite in SCB2
    lsr     #7,d1               ; bring level value to byte size (should be 8?)
    ori.w   #$0F00,d1           ; no horizontal shrinking, use level value for vertical shrinking
    move.w  d1,VRAM_RW

    move.w  d0,d1
    lsr     #8,d1               ; adjust Y sprite position for vertical centering
    lsr     #2,d1
    neg.w   d1
    addi.w  #128,d1             ; Y=128-((level>>8)/4)
    move.w  d1,d0

    move.l  d2,d3
    addi.w  #SCB3,d3
    move.w  d3,VRAM_ADDR        ; compute address for current sprite in SCB3
    jsr     SetSprY

    addq.w  #1,d2               ; next sprite
    dbra    d7,.processlevels

    lsr.l   #6,d4               ; make average out of all levels (/64)
    lsr.l   #3,d4               ; /8 to speed down animation
    move.w  PALOFFS,d0
    lsr.w   #8,d0
    andi.w  #$3E,d0             ; loop palette index every 32, 16bit values and align
    move.b  d0,PALINDX
    add.w   d4,PALOFFS          ; add to make the colors "scroll"

    jsr     BIOSF_VBL

    cmp.w   #$60,VBLC           ; need to at least skip the first $60 frames before starting raster interrupts (why ?)
    bne     noanim              ; is there a "CD busy" bit somewhere ?
    move    #$180*4,$3C000A	; interrupt every 4 raster lines
    move    #$0000,REG_HBLANKPOS
    move    #$00D0,REG_HBLANKCNT
noanim:

    addq.w  #1,VBLC
    movem.l (a7)+,d0-d7/a0-a6
    rte

RLI:
    move.w  #2,REG_IRQACK
    movem.l d0-d7/a0-a6,-(a7)

    clr.l   d0                  ; get and set current line backdrop color
    move.b  PALINDX,d0
    lea     PurpleSine,a0
    move.w  0(a0,d0.w),d1
    move.w  d1,PALETTES+(16*2*256)-2
    addq.b  #2,PALINDX          ; palette entries are 16bits
    andi.b  #$3F,PALINDX        ; loop every 32, 16bit values

    clr.l   d1
    move.w  REG_CDDALEFTL,d1
    lea     NEWLEFTLEVELS,a0    ; get the left channel's level
    move.w  INDX,d0
    move.w  d1,0(a0,d0.w)
    addq.w  #2,INDX             ; level entries are 16bits

    movem.l (a7)+,d0-d7/a0-a6
    rte

IRQ3:
    move.w  #1, REG_IRQACK
    rte

; ---- Main ----

Start:
    lea     $10F300,sp
    move.b  d0,REG_DIPSW        ; kick watchdog
    clr.b   FLAGS               ; disable animation for now
    clr.w   VBLC                ; clear frame counter
    move.w  #7,REG_IRQACK
    move.w  #$2000,sr           ; enable interrupts

    move.l  #(40*32)-1,d0       ; clears fix with tile $FF, palette 0
    move.w  #FIXMAP,VRAM_ADDR
    move.w  #$0001,VRAM_MOD
.clfix:
    move.w  #$00FF,VRAM_RW
    nop
    dbra    d0,.clfix

    lea     Palette,a0          ; load one palette
    lea     PALETTES,a1
    move.l  #16-1,d0
ldpalt:
    move.w  (a0)+,(a1)+
    dbra    d0,ldpalt

    lea     PALETTES+(16*2*256)-2,a0
    move.w  #BLACK,(a0)         ; sets default backdrop color (will be overwritten later)

    move.w  #SCB3,VRAM_ADDR     ; clear sprite maps
    move.w  #1,VRAM_MOD
    move.l  #380-1,d7
.clspr:
    move.w  #0,VRAM_RW
    dbra    d7,.clspr

    move.l  #1,d0               ; set up 64 sprites as vertical bars
    move.l  #64-1,d7
.setupsprites:
    move.l  d0,d1
    lsl.w   #6,d1               ; compute address for sprite in SCB1 (*2*32)
    move.w  d1,VRAM_ADDR        ; set 4 tiles in the sprite's map:
    move.w  #1,VRAM_RW          ; tile 1
    move.w  #0,VRAM_RW          ; palette 0, no special attributes
    move.w  #1,VRAM_RW          ; ...
    move.w  #0,VRAM_RW
    move.w  #1,VRAM_RW
    move.w  #0,VRAM_RW
    move.w  #1,VRAM_RW
    move.w  #0,VRAM_RW
    move.l  d0,d1
    add.w   #SCB2,d1            ; compute address for sprite in SCB2
    move.w  d1,VRAM_ADDR
    move.w  #$0F00,VRAM_RW      ; sprite is all flat
    move.l  d0,d1
    add.w   #SCB3,d1            ; compute address for sprite in SCB3
    move.w  d1,VRAM_ADDR
    move.w  #((496-80)<<7)+4,VRAM_RW     ; set initial Y position and height (4 tiles)
    move.l  d0,d1
    add.w   #SCB4,d1            ; compute address for sprite in SCB4
    move.w  d1,VRAM_ADDR
    move.l  d0,d1
    lsl.w   #2,d1               ; X=Sprite #*5 (sprites only contain a 4 pixel wide strip)
    add.w   d0,d1
    lsl.w   #7,d1
    move.w  d1,VRAM_RW
    addq.w  #1,d0               ; next sprite
    dbra    d7,.setupsprites

    move.b  d0,REG_DIPSW        ; kick watchdog (shouldn't be useful, interrupts are running)
    move.l  #Sprite,BIOS_UPSRC  ; upload a sprite tile in tile #1 (leave tile #0 blank)
    move.l  #128,BIOS_UPDEST    ; sprites tiles are 128 bytes
    move.b  #2,BIOS_UPZONE      ; SPR zone
    move.l  #128,BIOS_UPSIZE
    move.b  #0,BIOS_UPBANK      ; first bank
    jsr     BIOSF_UPLOAD

    clr.b   PALOFFS             ; init palette animation
    clr.b   PALINDX

    move.b  #1,FLAGS            ; allow animation

    clr.w   d0
    move.w  #$0502,d0           ; starts playing CDDA track 2 with no looping
    jsr     BIOSF_CDDACMD

.loop:
   bra     .loop

; ---- Routines ----

; Displays 16bit value pointed by A0. Set VRAM address before, thanks.
WriteHex:
    move.l  d0,-(a7)
    move.w  #$20,VRAM_MOD
    move.w  (a0),d0
    andi.w  #$F000,d0
    rol.w   #4,d0
    jsr     hexshift
    addi.w  #$30,d0             ; numbers start in fix tileset
    move.w  d0, VRAM_RW         ; MSB high nibble

    move.w  (a0),d0
    andi.w  #$0F00,d0
    lsr.w   #8,d0
    jsr     hexshift
    addi.w  #$30,d0
    move.w  d0, VRAM_RW         ; MSB low nibble

    move.w  (a0),d0
    andi.w  #$00F0,d0
    lsr.w   #4,d0
    jsr     hexshift
    addi.w  #$30,d0
    move.w  d0, VRAM_RW         ; LSB high nibble

    move.w  (a0),d0
    andi.w  #$000F,d0
    jsr     hexshift
    addi.w  #$30,d0
    move.w  d0, VRAM_RW
    move.l  (a7)+,d0            ; LSB low nibble
    rts

hexshift:
    cmp.b   #$A,d0              ; switch to letters if > 9
    blt.b   .r
    addq.w  #7,d0
.r
    rts

; Converts reversed 16bit 2's complement value to binary
; In/Out: D1
convert:
    move.l  d7,-(a7)
    moveq   #16-1,d7            ; 16 bits to reverse
    moveq   #0,d0
.rev:
    lsr	    #1,d1
    bcc	    .skip
    bset    d7,d0               ; write bits backwards
.skip:
    dbf	    d7,.rev
    move.l  d0,d1
    tst     d1
    bpl     .notneg             ; abs()
    not     d1
.notneg:
    move.l  (a7)+,d7
    rts

; Sets sprite Y position. Set VRAM address before, thanks.
SetSprY:
    neg.w   d0
    addi.w  #496,d0
    lsl.w   #7,d0
    add     #4,d0               ; ORs sprite height
    move.w  d0,VRAM_RW
    rts

; ---- Data ----

Palette:
    dc.w BLACK,WHITE,RED,GREEN  ; first 3 used for text
    dc.w BLACK,WHITE,RED,GREEN
    dc.w BLACK,WHITE,RED,GREEN
    dc.w BLACK,WHITE,RED,$0FA0  ; last one used for sprites (~orange)

Sprite:                         ; sprite tile filled with a 4 pixel wide bar on the right, color $F
    dc.b    $F0,$F0,$F0,$F0,$F0,$F0,$F0,$F0
    dc.b    $F0,$F0,$F0,$F0,$F0,$F0,$F0,$F0
    dc.b    $F0,$F0,$F0,$F0,$F0,$F0,$F0,$F0
    dc.b    $F0,$F0,$F0,$F0,$F0,$F0,$F0,$F0
    dc.b    $F0,$F0,$F0,$F0,$F0,$F0,$F0,$F0
    dc.b    $F0,$F0,$F0,$F0,$F0,$F0,$F0,$F0
    dc.b    $F0,$F0,$F0,$F0,$F0,$F0,$F0,$F0
    dc.b    $F0,$F0,$F0,$F0,$F0,$F0,$F0,$F0
    dc.b    $00,$00,$00,$00,$00,$00,$00,$00
    dc.b    $00,$00,$00,$00,$00,$00,$00,$00
    dc.b    $00,$00,$00,$00,$00,$00,$00,$00
    dc.b    $00,$00,$00,$00,$00,$00,$00,$00
    dc.b    $00,$00,$00,$00,$00,$00,$00,$00
    dc.b    $00,$00,$00,$00,$00,$00,$00,$00
    dc.b    $00,$00,$00,$00,$00,$00,$00,$00
    dc.b    $00,$00,$00,$00,$00,$00,$00,$00

PurpleSine:                     ; color values: black, purple, black
    dc.w    $0000,$0001,$0102,$0203,$0305,$0406,$0507,$0608
    dc.w    $0608,$0709,$0709,$080A,$080A,$080A,$090B,$090B
    dc.w    $090B,$090B,$080A,$080A,$080A,$0709,$0709,$0608
    dc.w    $0608,$0507,$0406,$0305,$0203,$0102,$0001,$0000
